@@ -1,181 +1,5 @@
 package Coro::ProcessPool;
 
-use strict;
-use warnings;
-use Carp;
-
-use AnyEvent;
-use Coro;
-use Coro::AnyEvent qw(sleep);
-use Coro::Channel;
-use Coro::ProcessPool::Process;
-use Coro::ProcessPool::Util qw(cpu_count);
-
-our $VERSION = '0.18';
-
-if ($^O eq 'MSWin32') {
-    die 'MSWin32 is not supported';
-}
-
-sub new {
-    my ($class, %param) = @_;
-    return bless {
-        max_procs => $param{max_procs} || cpu_count,
-        max_reqs  => $param{max_reqs}  || 0,
-        num_procs => 0,
-        procs     => Coro::Channel->new(),
-        pending   => {},
-    }, $class;
-}
-
-sub DESTROY {
-  my $self = shift;
-  $self->shutdown;
-}
-
-sub shutdown {
-    my $self = shift;
-    for (1 .. $self->{num_procs}) {
-        my $proc = $self->{procs}->get;
-        $proc->terminate;
-        --$self->{num_procs};
-    }
-}
-
-sub start_proc {
-    my $self = shift;
-    my $proc = Coro::ProcessPool::Process->new();
-    $proc->spawn;
-    ++$self->{num_procs};
-    return $proc;
-}
-
-sub kill_proc {
-    my ($self, $proc) = @_;
-    $proc->terminate;
-    --$self->{num_procs};
-}
-
-sub checkin_proc {
-    my ($self, $proc) = @_;
-    $self->{procs}->put($proc);
-}
-
-sub checkout_proc {
-    my ($self, $timeout) = @_;
-
-    # Start a new process if none are available and there are worker slots open
-    if ($self->{procs}->size == 0 && $self->{num_procs} < $self->{max_procs}) {
-        return $self->start_proc;
-    }
-
-    if (!defined $timeout) {
-        return $self->{procs}->get;
-    } else {
-        my $cv           = AnyEvent->condvar;
-        my $thread_timer = async { Coro::AnyEvent::sleep($timeout), $cv->send(0) };
-        my $thread_proc  = async { $cv->send($self->{procs}->get) };
-        my $proc         = $cv->recv;
-
-        if ($proc) {
-            $thread_timer->cancel;
-            return $proc;
-        } else {
-            croak 'timed out waiting for available process';
-        }
-    }
-}
-
-sub start_task {
-    my ($self, $f, $args, $timeout) = @_;
-    defined $f || croak 'expected CODE ref or task class (string) to execute';
-    $args ||= [];
-    ref $args eq 'ARRAY' || croak 'expected ARRAY ref of arguments';
-
-    my $proc = $self->checkout_proc($timeout);
-
-    # Restart the process if the worker is exhausted
-    my $thread;
-    if ($self->{max_reqs} > 0 && $proc->{processed} >= $self->{max_reqs}) {
-        $self->kill_proc($proc);
-        $proc = $self->start_proc;
-    }
-
-    # Send the task
-    my $msgid = $proc->send($f, $args);
-
-    # Note which process is handling this task
-    $self->{pending}{$msgid} = $proc;
-
-    # Replace process in the pool as soon as result is ready on the connection
-    async_pool {
-        my ($self, $proc) = @_;
-        $proc->readable;
-        $self->checkin_proc($proc);
-    } $self, $proc;
-
-    return $msgid;
-}
-
-sub collect_task {
-    my ($self, $msgid) = @_;
-    my $proc = $self->{pending}{$msgid} || croak 'msgid not found';
-    delete $self->{pending}{$msgid};
-    return $proc->recv($msgid);
-}
-
-sub process {
-    my $self  = shift;
-    my $msgid = $self->start_task(@_);
-    return $self->collect_task($msgid);
-}
-
-sub map {
-    my ($self, $f, @args) = @_;
-    my @deferred = map { $self->defer($f, [$_]) } @args;
-    return map { $_->() } @deferred;
-}
-
-sub defer {
-    my $self  = shift;
-    my $arr   = wantarray;
-    my $cv    = AnyEvent->condvar;
-    my $msgid = $self->start_task(@_);
-
-    async_pool {
-        my ($self, $arr, $msgid) = @_;
-
-        if ($arr) {
-            my @results = eval { $self->collect_task($msgid) };
-            $cv->croak($@) if $@;
-            $cv->send(@results);
-        } else {
-            my $result = eval { $self->collect_task($msgid) };
-            $cv->croak($@) if $@;
-            $cv->send($result);
-        }
-    } $self, $arr, $msgid;
-
-    return sub { $cv->recv };
-}
-
-sub queue {
-    my ($self, $f, $args, $k) = @_;
-    my $deferred = $self->defer($f, $args);
-
-    async_pool {
-        my ($deferred, $k) = @_;
-        if (ref $k && ref $k eq 'CODE') {
-            $k->($deferred->());
-        } else {
-            $deferred->();
-        }
-    } $deferred, $k;
-}
-
-1;
-__END__
-
 =head1 NAME
 
 Coro::ProcessPool - an asynchronous process pool
@@ -215,32 +39,254 @@ Coro::ProcessPool - an asynchronous process pool
 
 Processes tasks using a pool of external Perl processes.
 
+=cut
+
+use Moo;
+use Types::Standard qw(-types);
+use Carp;
+use AnyEvent;
+use Guard;
+use Coro;
+use Coro::AnyEvent qw(sleep);
+use Coro::Channel;
+use Coro::ProcessPool::Process;
+use Coro::ProcessPool::Util qw(cpu_count);
+use Coro::Semaphore;
+
+our $VERSION = '0.19';
+
+if ($^O eq 'MSWin32') {
+    die 'MSWin32 is not supported';
+}
+
+=head1 ATTRIBUTES
+
+=head2 max_procs
+
+The maximum number of processes to run within the process pool. Defaults
+to the number of CPUs on the ssytem.
+
+=cut
+
+has max_procs => (
+    is      => 'ro',
+    isa     => Int,
+    default => \&cpu_count,
+);
+
+=head2 max_reqs
+
+The maximum number of tasks a worker process may run before being terminated
+and replaced with a fresh process. This is useful for tasks that might leak
+memory over time.
+
+=cut
+
+has max_reqs => (
+    is      => 'ro',
+    isa     => Int,
+    default => 0,
+);
+
+=head1 PRIVATE ATTRIBUTES
+
+=head2 procs_lock
+
+Semaphore used to control access to the worker processes. Starts incremented
+to the number of processes (C<max_procs>).
+
+=cut
+
+has procs_lock => (
+    is      => 'lazy',
+    isa     => InstanceOf['Coro::Semaphore'],
+);
+
+sub _build_procs_lock {
+    my $self = shift;
+    return Coro::Semaphore->new($self->max_procs);
+}
+
+=head2 num_procs
+
+Running total of processes that are currently running.
+
+=cut
+
+has num_procs => (
+    is      => 'rw',
+    isa     => Int,
+    default => 0,
+);
+
+=head2 procs
+
+Array holding the L<Coro::Process> objects.
+
+=cut
+
+has procs => (
+    is      => 'ro',
+    isa     => ArrayRef[InstanceOf['Coro::Process']],
+    default => sub { [] },
+);
+
+=head2 inbox
+
+L<Coro::Channel> instanced used to queue tasks from the C<queue> method.
+
+=cut
+
+has inbox => (
+    is      => 'lazy',
+    isa     => InstanceOf['Coro::Channel'],
+);
+
+sub _build_inbox {
+  my $self = shift;
+  return Coro::Channel->new($self->max_procs);
+}
+
+=head2 inbox_worker
+
+L<Coro> thread monitoring the C<inbox>.
+
+=cut
+
+has inbox_worker => (
+    is  => 'lazy',
+    isa => InstanceOf['Coro'],
+);
+
+sub _build_inbox_worker {
+    async {
+        my $self = shift;
+
+        while (1) {
+            my $task = $self->inbox->get or last;
+            my ($f, $args, $on_success, $on_error) = @$task;
+
+            async_pool {
+                my ($self, $on_success, $on_error) = @_;
+                my $result = eval { $self->process($f, $args) };
+
+                if ($@) {
+                    my $error = $@;
+                    if (ref $on_error && ref $on_error eq 'CODE') {
+                        $on_error->($error);
+                    }
+                }
+                else {
+                    if (ref $on_success && ref $on_success eq 'CODE') {
+                        $on_success->($result);
+                    }
+                }
+            } $self, $on_success, $on_error;
+        }
+    } @_;
+}
+
+=head2 is_running
+
+Boolean which signals to the instance that the C<shutdown> method has been
+called.
+
+=cut
+
+has is_running => (
+    is      => 'rw',
+    isa     => Bool,
+    default => 1,
+);
+
 =head1 METHODS
 
-=head2 new
+=cut
 
-Creates a new process pool. Processes will be spawned as needed.
+sub DEMOLISH {
+    my $self = shift;
+    $self->shutdown;
+}
 
-=over
+sub BUILD {
+    my $self = shift;
+    $self->inbox_worker; # ensure the worker starts
+}
 
-=item max_procs
+sub start_proc {
+    my $self = shift;
+    my $proc = Coro::ProcessPool::Process->new();
+    ++$self->{num_procs};
+    return $proc;
+}
 
-This is the maximum number of child processes to maintain. If all processes are
-busy handling tasks, further calls to L<./process> will yield until a process
-becomes available. If not specified, defaults to the number of CPUs on the
-system.
+sub kill_proc {
+    my ($self, $proc) = @_;
+    $proc->shutdown;
+    --$self->{num_procs};
+}
 
-=item max_reqs
+sub checkin_proc {
+    my ($self, $proc) = @_;
 
-If this is a positive number (defaults to 0), child processes will be
-terminated and replaced after handling C<max_reqs> tasks. Choosing the correct
-value for C<max_reqs> is a tradeoff between the need to clear memory leaks in
-the child process and the time it takes to spawn a new process and import any
-packages used by client code.
+    if (!$self->is_running || ($self->max_reqs && $proc->messages_sent >= $self->max_reqs)) {
+        $self->kill_proc($proc);
+    } else {
+        unshift @{$self->procs}, $proc;
+    }
+}
 
-=back
+sub checkout_proc {
+    my $self = shift;
+    croak 'not running' unless $self->is_running;
 
-=head2 process($f, $args, $timeout)
+    my $proc;
+
+    # Start a new process if none are available and there are worker slots open
+    if ($self->capacity == 0 && $self->num_procs < $self->max_procs) {
+        $proc = $self->start_proc;
+    } else {
+        $proc = shift @{$self->procs};
+    }
+
+    return $proc;
+}
+
+=head2 capacity
+
+Returns the number of free worker processes.
+
+=cut
+
+sub capacity {
+    my $self = shift;
+    return scalar(@{$self->procs});
+}
+
+=head2 shutdown
+
+Shuts down all processes and resets state on the process pool. After calling
+this method, the pool is effectively in a new state and may be used normally.
+
+=cut
+
+sub shutdown {
+    my $self = shift;
+
+    $self->is_running(0);
+
+    $self->inbox->shutdown;
+    $self->inbox_worker->join;
+
+    my $count = $self->num_procs or return;
+    for (1 .. $count) {
+        my $proc = shift @{$self->procs};
+        $proc->shutdown(5);
+        --$self->{num_procs};
+    }
+}
+
+=head2 process($f, $args)
 
 Processes code ref C<$f> in a child process from the pool. If C<$args> is
 provided, it is an array ref of arguments that will be passed to C<$f>. Returns
@@ -256,12 +302,19 @@ busy, this method will block until one becomes available. Processes are spawned
 as needed, up to C<max_procs>, from this method. Also note that the use of
 C<max_reqs> can cause this method to yield while a new process is spawned.
 
-A timeout may be optionally specified in fractional seconds. If specified,
-C<$timeout> will cause C<process> to croak if C<$timeout> seconds pass an no
-process becomes available to handle the task.
+=cut
 
-Note that the timeout only applies to the time it takes to acquire an available
-process. It does not watch the time it takes to perform the task.
+sub process {
+    my ($self, $f, $args) = @_;
+
+    my $guard = $self->procs_lock->guard;
+
+    my $proc = $self->checkout_proc;
+    scope_guard { $self->checkin_proc($proc) };
+
+    my $msgid = $proc->send($f, $args);
+    return $proc->recv($msgid);
+}
 
 =head2 map($f, @args)
 
@@ -271,6 +324,14 @@ guaranteed, the results are guaranteed to be in the same order as C<@args>,
 even if the result of calling C<$f> returns a list itself (in which case, the
 results of that calcuation is flattened into the list returned by C<map>.
 
+=cut
+
+sub map {
+    my ($self, $f, @args) = @_;
+    my @deferred = map { $self->defer($f, [$_]) } @args;
+    return map { $_->() } @deferred;
+}
+
 =head2 defer($f, $args)
 
 Similar to L<./process>, but returns immediately. The return value is a code
@@ -279,6 +340,22 @@ reference that, when called, returns the results of calling C<$f->(@$args)>.
     my $deferred = $pool->defer($coderef, [ $x, $y, $z ]);
     my $result   = $deferred->();
 
+=cut
+
+sub defer {
+    my $self = shift;
+    my $cv   = AnyEvent->condvar;
+
+    async_pool {
+        my ($self, $cv, @args) = @_;
+        my $result = eval { $self->process(@args) };
+        $cv->croak($@) if $@;
+        $cv->send($result);
+    } $self, $cv, @_;
+
+    return sub { $cv->recv };
+}
+
 =head2 queue($f, $args, $callback)
 
 Queues the execution of C<$f->(@$args)> and returns immediately. If
@@ -303,10 +380,14 @@ variable.
     $pool->queue($doubler_function, [21], make_callback(21));
     $cv->recv;
 
-=head2 shutdown
+=cut
 
-Shuts down all processes and resets state on the process pool. After calling
-this method, the pool is effectively in a new state and may be used normally.
+sub queue {
+    my ($self, $f, $args, $on_success, $on_error) = @_;
+    $self->inbox->put([$f, $args, $on_success, $on_error]);
+}
+
+=back
 
 =head1 A NOTE ABOUT IMPORTS AND CLOSURES
 
@@ -412,8 +493,6 @@ process pool on Windows:
 
 =item L<Win32::Pipe>
 
-=back
-
 =head1 AUTHOR
 
 Jeff Ober <jeffober@gmail.com>
@@ -421,3 +500,7 @@ Jeff Ober <jeffober@gmail.com>
 =head1 LICENSE
 
 BSD License
+
+=cut
+
+1;
